"""Tests for the security monitoring module."""

from __future__ import annotations

import json
import tempfile
from datetime import UTC, datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from gaol.security.apparmor import (
    DEFAULT_CAPABILITIES_ALLOW,
    DEFAULT_CAPABILITIES_DENY,
    DEFAULT_FILE_DENY,
    DEFAULT_FILE_READ,
    AppArmorProfileGenerator,
)
from gaol.security.client import (
    FalcoClient,
    FalcoEventBuffer,
)
from gaol.security.exceptions import (
    FalcoConnectionError,
    FalcoNotInstalledError,
    FalcoNotRunningError,
    RuleExistsError,
    RuleNotFoundError,
    RuleValidationError,
    SecurityError,
    UnsupportedDistroError,
)
from gaol.security.installer import (
    FalcoInstaller,
    PackageManager,
)
from gaol.security.models import (
    AppArmorMode,
    AppArmorProfile,
    ContainerInfo,
    EventSummary,
    FalcoPriority,
    FalcoStatus,
    RuleList,
    SeccompAction,
    SeccompProfile,
    SeccompSyscall,
    SecurityEvent,
    SecurityRule,
)
from gaol.security.rules import (
    RuleManager,
)
from gaol.security.seccomp import (
    BASELINE_SYSCALLS,
    SeccompProfileGenerator,
)

# =============================================================================
# Model Tests
# =============================================================================


class TestFalcoPriority:
    """Tests for FalcoPriority enum."""

    def test_from_string_valid(self) -> None:
        """Test parsing valid priority strings."""
        assert FalcoPriority.from_string("Emergency") == FalcoPriority.EMERGENCY
        assert FalcoPriority.from_string("warning") == FalcoPriority.WARNING
        assert FalcoPriority.from_string("DEBUG") == FalcoPriority.DEBUG

    def test_from_string_invalid(self) -> None:
        """Test parsing invalid priority strings raises ValueError."""
        with pytest.raises(ValueError, match="Unknown priority"):
            FalcoPriority.from_string("invalid")

    def test_priority_comparison(self) -> None:
        """Test priority comparison (Emergency is highest)."""
        assert FalcoPriority.EMERGENCY < FalcoPriority.ALERT
        assert FalcoPriority.ALERT < FalcoPriority.CRITICAL
        assert FalcoPriority.WARNING < FalcoPriority.DEBUG
        assert FalcoPriority.EMERGENCY <= FalcoPriority.EMERGENCY
        assert FalcoPriority.WARNING <= FalcoPriority.DEBUG

    def test_priority_comparison_with_non_priority(self) -> None:
        """Test comparison with non-FalcoPriority returns NotImplemented."""
        assert FalcoPriority.WARNING.__lt__("string") == NotImplemented
        assert FalcoPriority.WARNING.__le__(123) == NotImplemented


class TestContainerInfo:
    """Tests for ContainerInfo model."""

    def test_container_info_creation(self) -> None:
        """Test creating container info."""
        info = ContainerInfo(
            id="abc123",
            name="my-container",
            image="nginx",
            image_tag="latest",
        )
        assert info.id == "abc123"
        assert info.name == "my-container"
        assert info.image == "nginx"
        assert info.image_tag == "latest"

    def test_container_info_defaults(self) -> None:
        """Test container info defaults to None."""
        info = ContainerInfo()
        assert info.id is None
        assert info.name is None
        assert info.image is None


class TestSecurityEvent:
    """Tests for SecurityEvent model."""

    def test_security_event_creation(self) -> None:
        """Test creating a security event."""
        event = SecurityEvent(
            uuid="test-uuid",
            timestamp=datetime.now(UTC),
            rule="test_rule",
            priority=FalcoPriority.WARNING,
            output="Test output message",
        )
        assert event.uuid == "test-uuid"
        assert event.rule == "test_rule"
        assert event.priority == FalcoPriority.WARNING

    def test_matches_container(self) -> None:
        """Test container name pattern matching."""
        event = SecurityEvent(
            uuid="test",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
            container=ContainerInfo(name="my-app-container"),
        )
        assert event.matches_container("my-app.*")
        assert event.matches_container("my-app-container")
        assert not event.matches_container("other-.*")

    def test_matches_container_no_container(self) -> None:
        """Test matching when no container info."""
        event = SecurityEvent(
            uuid="test",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
        )
        assert not event.matches_container(".*")

    def test_is_at_least_priority(self) -> None:
        """Test priority level checking."""
        event = SecurityEvent(
            uuid="test",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
        )
        # WARNING is at least WARNING
        assert event.is_at_least_priority(FalcoPriority.WARNING)
        # WARNING is at least DEBUG (lower priority)
        assert event.is_at_least_priority(FalcoPriority.DEBUG)
        # WARNING is NOT at least ERROR (higher priority)
        assert not event.is_at_least_priority(FalcoPriority.ERROR)


class TestSecurityRule:
    """Tests for SecurityRule model."""

    def test_security_rule_creation(self) -> None:
        """Test creating a security rule."""
        rule = SecurityRule(
            name="test_rule",
            description="A test rule",
            condition="container and proc.name=bash",
            output="Shell spawned",
            priority=FalcoPriority.WARNING,
            tags=["container", "shell"],
        )
        assert rule.name == "test_rule"
        assert rule.enabled is True
        assert rule.gaol_managed is True

    def test_to_falco_yaml(self) -> None:
        """Test converting rule to Falco YAML format."""
        rule = SecurityRule(
            name="test_rule",
            description="Test description",
            condition="container",
            output="Output message",
            priority=FalcoPriority.CRITICAL,
            tags=["test"],
        )
        yaml_dict = rule.to_falco_yaml()

        assert yaml_dict["rule"] == "test_rule"
        assert yaml_dict["desc"] == "Test description"
        assert yaml_dict["condition"] == "container"
        assert yaml_dict["output"] == "Output message"
        assert yaml_dict["priority"] == "Critical"
        assert yaml_dict["tags"] == ["test"]
        assert yaml_dict["enabled"] is True


class TestRuleList:
    """Tests for RuleList model."""

    def test_rule_list_to_falco_yaml(self) -> None:
        """Test converting rule list to Falco YAML format."""
        rule_list = RuleList(
            macros={"container": "container.id != host"},
            lists={"shells": ["bash", "sh", "zsh"]},
            rules=[
                SecurityRule(
                    name="test_rule",
                    condition="container",
                    output="test",
                    priority=FalcoPriority.WARNING,
                )
            ],
        )
        yaml_items = rule_list.to_falco_yaml()

        # Should have: engine version, macro, list, rule
        assert len(yaml_items) == 4
        assert yaml_items[0] == {"required_engine_version": "0.26.0"}
        assert yaml_items[1] == {"macro": "container", "condition": "container.id != host"}
        assert yaml_items[2] == {"list": "shells", "items": ["bash", "sh", "zsh"]}
        assert yaml_items[3]["rule"] == "test_rule"


class TestSeccompProfile:
    """Tests for SeccompProfile model."""

    def test_seccomp_profile_creation(self) -> None:
        """Test creating a seccomp profile."""
        profile = SeccompProfile(
            name="test-profile",
            defaultAction=SeccompAction.SCMP_ACT_ERRNO,
            syscalls=[
                SeccompSyscall(
                    names=["read", "write"],
                    action=SeccompAction.SCMP_ACT_ALLOW,
                )
            ],
        )
        assert profile.name == "test-profile"
        assert profile.defaultAction == SeccompAction.SCMP_ACT_ERRNO
        assert len(profile.syscalls) == 1

    def test_to_oci_json(self) -> None:
        """Test converting to OCI JSON format."""
        profile = SeccompProfile(
            defaultAction=SeccompAction.SCMP_ACT_ERRNO,
            syscalls=[
                SeccompSyscall(
                    names=["read", "write"],
                    action=SeccompAction.SCMP_ACT_ALLOW,
                )
            ],
        )
        oci = profile.to_oci_json()

        assert oci["defaultAction"] == "SCMP_ACT_ERRNO"
        assert len(oci["syscalls"]) == 1
        assert oci["syscalls"][0]["names"] == ["read", "write"]
        assert oci["syscalls"][0]["action"] == "SCMP_ACT_ALLOW"

    def test_add_syscall_new_action(self) -> None:
        """Test adding syscalls with new action."""
        profile = SeccompProfile()
        profile.add_syscall(["read", "write"], SeccompAction.SCMP_ACT_ALLOW)

        assert len(profile.syscalls) == 1
        assert "read" in profile.syscalls[0].names
        assert "write" in profile.syscalls[0].names

    def test_add_syscall_existing_action(self) -> None:
        """Test adding syscalls to existing action group."""
        profile = SeccompProfile(
            syscalls=[
                SeccompSyscall(names=["read"], action=SeccompAction.SCMP_ACT_ALLOW)
            ]
        )
        profile.add_syscall(["write"], SeccompAction.SCMP_ACT_ALLOW)

        assert len(profile.syscalls) == 1
        assert "read" in profile.syscalls[0].names
        assert "write" in profile.syscalls[0].names


class TestAppArmorProfile:
    """Tests for AppArmorProfile model."""

    def test_apparmor_profile_creation(self) -> None:
        """Test creating an AppArmor profile."""
        profile = AppArmorProfile(
            name="test-profile",
            mode=AppArmorMode.ENFORCE,
            capabilities_allow=["chown", "setuid"],
            file_read=["/bin/**"],
        )
        assert profile.name == "test-profile"
        assert profile.mode == AppArmorMode.ENFORCE
        assert "chown" in profile.capabilities_allow

    def test_to_apparmor_text(self) -> None:
        """Test generating AppArmor profile text."""
        profile = AppArmorProfile(
            name="test_profile",
            capabilities_allow=["chown"],
            capabilities_deny=["sys_admin"],
            file_read=["/bin/**"],
            file_write=["/tmp/**"],
            network_allow=["inet stream"],
        )
        text = profile.to_apparmor_text()

        assert "profile test_profile" in text
        assert "capability chown," in text
        assert "deny capability sys_admin," in text
        assert "/bin/** r," in text
        assert "/tmp/** w," in text
        assert "network inet stream," in text

    def test_capabilities_lowercase(self) -> None:
        """Test capabilities are lowercased."""
        profile = AppArmorProfile(
            name="test",
            capabilities_allow=["CHOWN", "SetUID"],
        )
        assert profile.capabilities_allow == ["chown", "setuid"]


class TestFalcoStatus:
    """Tests for FalcoStatus model."""

    def test_falco_status_defaults(self) -> None:
        """Test FalcoStatus default values."""
        status = FalcoStatus()
        assert status.installed is False
        assert status.running is False
        assert status.version is None
        assert status.grpc_enabled is False


class TestEventSummary:
    """Tests for EventSummary model."""

    def test_event_summary_defaults(self) -> None:
        """Test EventSummary default values."""
        summary = EventSummary()
        assert summary.total_events == 0
        assert summary.events_by_priority == {}
        assert summary.events_by_rule == {}
        assert summary.events_by_container == {}


# =============================================================================
# Exception Tests
# =============================================================================


class TestSecurityExceptions:
    """Tests for security exceptions."""

    def test_security_error_base(self) -> None:
        """Test SecurityError is base exception."""
        with pytest.raises(SecurityError):
            raise FalcoConnectionError("test")

    def test_unsupported_distro_error(self) -> None:
        """Test UnsupportedDistroError message."""
        error = UnsupportedDistroError("arch")
        assert "arch" in str(error)
        assert error.distro == "arch"

    def test_rule_validation_error(self) -> None:
        """Test RuleValidationError with errors list."""
        error = RuleValidationError("test_rule", ["error1", "error2"])
        assert "test_rule" in str(error)
        assert error.name == "test_rule"
        assert len(error.errors) == 2

    def test_rule_not_found_error(self) -> None:
        """Test RuleNotFoundError."""
        error = RuleNotFoundError("missing_rule")
        assert "missing_rule" in str(error)
        assert error.name == "missing_rule"

    def test_rule_exists_error(self) -> None:
        """Test RuleExistsError."""
        error = RuleExistsError("existing_rule")
        assert "existing_rule" in str(error)
        assert error.name == "existing_rule"


# =============================================================================
# FalcoClient Tests
# =============================================================================


class TestFalcoClient:
    """Tests for FalcoClient."""

    def test_client_creation(self) -> None:
        """Test creating a client with default settings."""
        client = FalcoClient()
        assert client.is_connected() is False
        assert "unix://" in client.endpoint

    def test_client_with_grpc_address(self) -> None:
        """Test creating client with gRPC address."""
        client = FalcoClient(grpc_address="localhost:5060")
        assert client.endpoint == "localhost:5060"

    def test_client_with_socket_path(self) -> None:
        """Test creating client with custom socket path."""
        client = FalcoClient(socket_path="/custom/socket.sock")
        assert "/custom/socket.sock" in client.endpoint

    @patch.object(FalcoClient, "_is_falco_installed")
    def test_connect_not_installed(self, mock_installed: MagicMock) -> None:
        """Test connect raises when Falco not installed."""
        mock_installed.return_value = False
        client = FalcoClient()

        with pytest.raises(FalcoNotInstalledError):
            client.connect()

    @patch.object(FalcoClient, "_is_falco_installed")
    @patch.object(FalcoClient, "_is_falco_running")
    def test_connect_not_running(
        self, mock_running: MagicMock, mock_installed: MagicMock
    ) -> None:
        """Test connect raises when Falco not running."""
        mock_installed.return_value = True
        mock_running.return_value = False
        client = FalcoClient()

        with pytest.raises(FalcoNotRunningError):
            client.connect()

    def test_disconnect(self) -> None:
        """Test disconnecting client."""
        client = FalcoClient()
        client._connected = True
        client.disconnect()
        assert client.is_connected() is False

    def test_get_status_not_installed(self) -> None:
        """Test get_status when Falco not installed."""
        with patch.object(FalcoClient, "_is_falco_installed", return_value=False):
            client = FalcoClient()
            status = client.get_status()
            assert status.installed is False
            assert status.running is False

    def test_parse_event_line_json(self) -> None:
        """Test parsing JSON event line."""
        client = FalcoClient()
        json_line = json.dumps({
            "uuid": "test-uuid",
            "time": "2024-01-01T00:00:00Z",
            "rule": "test_rule",
            "priority": "Warning",
            "output": "Test output",
            "source": "syscall",
        })

        event = client._parse_event_line(json_line)
        assert event is not None
        assert event.uuid == "test-uuid"
        assert event.rule == "test_rule"
        assert event.priority == FalcoPriority.WARNING

    def test_parse_event_line_invalid(self) -> None:
        """Test parsing invalid event line is handled."""
        client = FalcoClient()
        # The parser is lenient and may return None or attempt to parse
        # We just verify it doesn't crash on invalid input
        result = client._parse_event_line("not a valid event")
        # Result can be None or a partially parsed event
        assert result is None or isinstance(result, SecurityEvent)


class TestFalcoEventBuffer:
    """Tests for FalcoEventBuffer."""

    def test_buffer_creation(self) -> None:
        """Test creating an event buffer."""
        buffer = FalcoEventBuffer(max_size=100)
        assert len(buffer) == 0

    def test_add_event(self) -> None:
        """Test adding events to buffer."""
        buffer = FalcoEventBuffer()
        event = SecurityEvent(
            uuid="test",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
        )
        buffer.add(event)
        assert len(buffer) == 1

    def test_buffer_max_size(self) -> None:
        """Test buffer respects max size."""
        buffer = FalcoEventBuffer(max_size=2)

        for i in range(5):
            event = SecurityEvent(
                uuid=f"test-{i}",
                timestamp=datetime.now(UTC),
                rule="test",
                priority=FalcoPriority.WARNING,
                output=f"test {i}",
            )
            buffer.add(event)

        assert len(buffer) == 2

    def test_get_recent(self) -> None:
        """Test getting recent events."""
        buffer = FalcoEventBuffer()

        for i in range(5):
            event = SecurityEvent(
                uuid=f"test-{i}",
                timestamp=datetime.now(UTC),
                rule="test",
                priority=FalcoPriority.WARNING,
                output=f"test {i}",
            )
            buffer.add(event)

        recent = buffer.get_recent(count=3)
        assert len(recent) == 3
        # Most recent first
        assert recent[0].uuid == "test-4"

    def test_get_recent_with_priority_filter(self) -> None:
        """Test filtering events by priority."""
        buffer = FalcoEventBuffer()

        buffer.add(SecurityEvent(
            uuid="low",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.DEBUG,
            output="debug",
        ))
        buffer.add(SecurityEvent(
            uuid="high",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.CRITICAL,
            output="critical",
        ))

        recent = buffer.get_recent(min_priority=FalcoPriority.WARNING)
        assert len(recent) == 1
        assert recent[0].uuid == "high"

    def test_get_recent_with_container_filter(self) -> None:
        """Test filtering events by container."""
        buffer = FalcoEventBuffer()

        buffer.add(SecurityEvent(
            uuid="match",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
            container=ContainerInfo(name="my-app"),
        ))
        buffer.add(SecurityEvent(
            uuid="no-match",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
            container=ContainerInfo(name="other-app"),
        ))

        recent = buffer.get_recent(container="my-.*")
        assert len(recent) == 1
        assert recent[0].uuid == "match"

    def test_get_summary(self) -> None:
        """Test getting event summary."""
        buffer = FalcoEventBuffer()

        buffer.add(SecurityEvent(
            uuid="1",
            timestamp=datetime.now(UTC),
            rule="rule_a",
            priority=FalcoPriority.WARNING,
            output="test",
        ))
        buffer.add(SecurityEvent(
            uuid="2",
            timestamp=datetime.now(UTC),
            rule="rule_a",
            priority=FalcoPriority.WARNING,
            output="test",
        ))
        buffer.add(SecurityEvent(
            uuid="3",
            timestamp=datetime.now(UTC),
            rule="rule_b",
            priority=FalcoPriority.CRITICAL,
            output="test",
        ))

        summary = buffer.get_summary()
        assert summary.total_events == 3
        assert summary.events_by_rule["rule_a"] == 2
        assert summary.events_by_rule["rule_b"] == 1
        assert summary.events_by_priority["Warning"] == 2
        assert summary.events_by_priority["Critical"] == 1

    def test_clear(self) -> None:
        """Test clearing buffer."""
        buffer = FalcoEventBuffer()
        buffer.add(SecurityEvent(
            uuid="test",
            timestamp=datetime.now(UTC),
            rule="test",
            priority=FalcoPriority.WARNING,
            output="test",
        ))
        assert len(buffer) == 1

        buffer.clear()
        assert len(buffer) == 0


# =============================================================================
# FalcoInstaller Tests
# =============================================================================


class TestFalcoInstaller:
    """Tests for FalcoInstaller."""

    def test_detect_package_manager_unknown(self) -> None:
        """Test detecting package manager on unknown system raises error."""
        with patch("shutil.which", return_value=None):
            installer = FalcoInstaller()
            with pytest.raises(UnsupportedDistroError):
                installer.detect_package_manager()

    def test_detect_package_manager_apt(self) -> None:
        """Test detecting apt package manager."""
        def mock_which(cmd: str) -> str | None:
            return "/usr/bin/apt-get" if cmd == "apt-get" else None

        with patch("shutil.which", side_effect=mock_which):
            installer = FalcoInstaller()
            pm = installer.detect_package_manager()
            assert pm == PackageManager.APT

    def test_detect_package_manager_dnf(self) -> None:
        """Test detecting dnf package manager."""
        def mock_which(cmd: str) -> str | None:
            return "/usr/bin/dnf" if cmd == "dnf" else None

        with patch("shutil.which", side_effect=mock_which):
            installer = FalcoInstaller()
            pm = installer.detect_package_manager()
            assert pm == PackageManager.DNF

    def test_is_falco_installed_true(self) -> None:
        """Test checking if Falco is installed."""
        with patch("shutil.which", return_value="/usr/bin/falco"):
            installer = FalcoInstaller()
            assert installer.is_falco_installed() is True

    def test_is_falco_installed_false(self) -> None:
        """Test checking when Falco is not installed."""
        with patch("shutil.which", return_value=None):
            installer = FalcoInstaller()
            assert installer.is_falco_installed() is False

    @patch("subprocess.run")
    def test_get_installed_version(self, mock_run: MagicMock) -> None:
        """Test getting installed Falco version."""
        mock_run.return_value = MagicMock(
            returncode=0,
            stdout="Falco 0.37.1\nDriver version: abc123",
        )
        with patch("shutil.which", return_value="/usr/bin/falco"):
            installer = FalcoInstaller()
            version = installer.get_installed_version()
            assert version == "0.37.1"

    def test_install_unsupported_distro(self) -> None:
        """Test install raises on unsupported distro."""
        with (
            patch("shutil.which", return_value=None),
            patch("os.geteuid", return_value=0),  # Pretend we're root
        ):
            installer = FalcoInstaller()
            with pytest.raises(UnsupportedDistroError):
                installer.install()

    def test_get_status(self) -> None:
        """Test getting Falco status."""
        with patch("shutil.which", return_value=None):
            installer = FalcoInstaller()
            status = installer.get_status()
            assert status.installed is False


# =============================================================================
# RuleManager Tests
# =============================================================================


class TestRuleManager:
    """Tests for RuleManager."""

    def test_rule_manager_creation(self) -> None:
        """Test creating rule manager with custom dirs."""
        with tempfile.TemporaryDirectory() as tmpdir:
            rules_dir = Path(tmpdir) / "rules.d"
            gaol_dir = rules_dir / "gaol"

            manager = RuleManager(
                rules_dir=rules_dir,
                gaol_dir=gaol_dir,
            )
            assert manager.rules_dir == rules_dir
            assert manager.gaol_dir == gaol_dir

    def test_create_rule(self) -> None:
        """Test creating a new rule."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            rule = SecurityRule(
                name="test_rule",
                condition="container",
                output="test output",
                priority=FalcoPriority.WARNING,
            )

            path = manager.create_rule(rule)
            assert path.exists()
            assert "test_rule" in path.name

    def test_create_rule_already_exists(self) -> None:
        """Test creating duplicate rule raises error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            rule = SecurityRule(
                name="test_rule",
                condition="container",
                output="test",
                priority=FalcoPriority.WARNING,
            )

            manager.create_rule(rule)

            with pytest.raises(RuleExistsError):
                manager.create_rule(rule)

    def test_get_rule(self) -> None:
        """Test getting a rule by name."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            rule = SecurityRule(
                name="test_rule",
                description="Test description",
                condition="container",
                output="test",
                priority=FalcoPriority.WARNING,
            )
            manager.create_rule(rule)

            retrieved = manager.get_rule("test_rule")
            assert retrieved.name == "test_rule"
            assert retrieved.description == "Test description"

    def test_get_rule_not_found(self) -> None:
        """Test getting non-existent rule raises error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            with pytest.raises(RuleNotFoundError):
                manager.get_rule("nonexistent")

    def test_list_rules(self) -> None:
        """Test listing rules."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            for i in range(3):
                rule = SecurityRule(
                    name=f"rule_{i}",
                    condition="container",
                    output="test",
                    priority=FalcoPriority.WARNING,
                )
                manager.create_rule(rule)

            rules = manager.list_rules(include_builtin=False)
            assert len(rules) == 3

    def test_delete_rule(self) -> None:
        """Test deleting a rule."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            rule = SecurityRule(
                name="test_rule",
                condition="container",
                output="test",
                priority=FalcoPriority.WARNING,
            )
            manager.create_rule(rule)

            manager.delete_rule("test_rule")

            with pytest.raises(RuleNotFoundError):
                manager.get_rule("test_rule")

    def test_enable_disable_rule(self) -> None:
        """Test enabling and disabling rules."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            rule = SecurityRule(
                name="test_rule",
                condition="container",
                output="test",
                priority=FalcoPriority.WARNING,
                enabled=True,
            )
            manager.create_rule(rule)

            manager.disable_rule("test_rule")
            updated = manager.get_rule("test_rule")
            assert updated.enabled is False

            manager.enable_rule("test_rule")
            updated = manager.get_rule("test_rule")
            assert updated.enabled is True

    def test_validate_rule_valid(self) -> None:
        """Test validating a valid rule."""
        manager = RuleManager()
        rule = SecurityRule(
            name="valid_rule",
            condition="container and proc.name=bash",
            output="Shell spawned",
            priority=FalcoPriority.WARNING,
        )
        errors = manager.validate_rule(rule)
        assert len(errors) == 0

    def test_validate_rule_missing_condition_and_output(self) -> None:
        """Test validating rule with empty condition and output."""
        manager = RuleManager()
        # Note: Pydantic validates name format at model creation
        # So we test with a valid name but empty condition/output
        rule = SecurityRule(
            name="valid_name",
            condition="",
            output="",
            priority=FalcoPriority.WARNING,
        )
        errors = manager.validate_rule(rule)
        assert "Rule condition is required" in errors
        assert "Rule output is required" in errors

    def test_validate_rule_invalid_name_raises_at_creation(self) -> None:
        """Test that invalid name raises Pydantic validation error."""
        from pydantic import ValidationError

        # Pydantic validates name pattern at model creation
        with pytest.raises(ValidationError):
            SecurityRule(
                name="invalid-name-with-dashes",
                condition="container",
                output="test",
                priority=FalcoPriority.WARNING,
            )

    def test_validate_rule_unbalanced_parens(self) -> None:
        """Test validating rule with unbalanced parentheses."""
        manager = RuleManager()
        rule = SecurityRule(
            name="test_rule",
            condition="(container and (proc.name=bash)",
            output="test",
            priority=FalcoPriority.WARNING,
        )
        errors = manager.validate_rule(rule)
        assert any("parentheses" in e.lower() for e in errors)

    def test_import_export_rules(self) -> None:
        """Test importing and exporting rules."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            # Create some rules
            for i in range(2):
                rule = SecurityRule(
                    name=f"rule_{i}",
                    condition="container",
                    output="test",
                    priority=FalcoPriority.WARNING,
                )
                manager.create_rule(rule)

            # Export
            export_path = Path(tmpdir) / "exported.yaml"
            count = manager.export_rules(export_path)
            assert count == 2
            assert export_path.exists()

            # Create new manager and import
            manager2 = RuleManager(
                rules_dir=Path(tmpdir) / "new",
                gaol_dir=Path(tmpdir) / "new" / "gaol",
            )
            imported = manager2.import_rules(export_path)
            assert imported == 2


# =============================================================================
# SeccompProfileGenerator Tests
# =============================================================================


class TestSeccompProfileGenerator:
    """Tests for SeccompProfileGenerator."""

    def test_generator_creation(self) -> None:
        """Test creating generator with custom dir."""
        with tempfile.TemporaryDirectory() as tmpdir:
            generator = SeccompProfileGenerator(profiles_dir=Path(tmpdir))
            assert generator.profiles_dir == Path(tmpdir)

    def test_generate_default(self) -> None:
        """Test generating default profile."""
        generator = SeccompProfileGenerator()
        profile = generator.generate_default("test-container")

        assert profile.name == "test-container"
        assert profile.defaultAction == SeccompAction.SCMP_ACT_ERRNO
        assert len(profile.syscalls) > 0
        assert profile.generated_from == "gaol-default"

    def test_generate_restrictive(self) -> None:
        """Test generating restrictive profile."""
        generator = SeccompProfileGenerator()
        profile = generator.generate_restrictive("test-container")

        assert profile.name == "test-container"
        assert profile.defaultAction == SeccompAction.SCMP_ACT_KILL_PROCESS
        assert profile.generated_from == "gaol-restrictive"

    def test_save_and_load_profile(self) -> None:
        """Test saving and loading a profile."""
        with tempfile.TemporaryDirectory() as tmpdir:
            generator = SeccompProfileGenerator(profiles_dir=Path(tmpdir))

            profile = generator.generate_default("test")
            saved_path = generator.save_profile(profile)

            assert saved_path.exists()

            loaded = generator.load_profile(saved_path)
            assert loaded.name == "test"
            assert loaded.defaultAction == SeccompAction.SCMP_ACT_ERRNO

    def test_load_profile_not_found(self) -> None:
        """Test loading non-existent profile raises error."""
        generator = SeccompProfileGenerator()

        with pytest.raises(FileNotFoundError):
            generator.load_profile(Path("/nonexistent/profile.json"))

    def test_list_profiles(self) -> None:
        """Test listing saved profiles."""
        with tempfile.TemporaryDirectory() as tmpdir:
            generator = SeccompProfileGenerator(profiles_dir=Path(tmpdir))

            # Initially empty
            assert len(generator.list_profiles()) == 0

            # Save some profiles
            for i in range(3):
                profile = generator.generate_default(f"profile-{i}")
                generator.save_profile(profile)

            profiles = generator.list_profiles()
            assert len(profiles) == 3

    def test_baseline_syscalls_not_empty(self) -> None:
        """Test that baseline syscalls list is populated."""
        assert len(BASELINE_SYSCALLS) > 50
        assert "read" in BASELINE_SYSCALLS
        assert "write" in BASELINE_SYSCALLS
        assert "open" in BASELINE_SYSCALLS


# =============================================================================
# AppArmorProfileGenerator Tests
# =============================================================================


class TestAppArmorProfileGenerator:
    """Tests for AppArmorProfileGenerator."""

    def test_generator_creation(self) -> None:
        """Test creating generator."""
        with tempfile.TemporaryDirectory() as tmpdir:
            generator = AppArmorProfileGenerator(profiles_dir=Path(tmpdir))
            assert generator.profiles_dir == Path(tmpdir)

    def test_generate_default(self) -> None:
        """Test generating default profile."""
        generator = AppArmorProfileGenerator()
        profile = generator.generate_default("my-container")

        assert profile.name == "my-container"
        assert profile.mode == AppArmorMode.ENFORCE
        assert len(profile.capabilities_allow) > 0
        assert len(profile.file_read) > 0

    def test_generate_restrictive(self) -> None:
        """Test generating restrictive profile."""
        generator = AppArmorProfileGenerator()
        profile = generator.generate_restrictive("my-container")

        assert profile.name == "my-container"
        # Restrictive profile should have fewer capabilities
        default_profile = generator.generate_default("other")
        assert len(profile.capabilities_allow) < len(default_profile.capabilities_allow)

    def test_sanitize_profile_name(self) -> None:
        """Test profile name sanitization."""
        generator = AppArmorProfileGenerator()

        assert generator._sanitize_profile_name("valid_name") == "valid_name"
        assert generator._sanitize_profile_name("name-with-dashes") == "name-with-dashes"
        assert generator._sanitize_profile_name("name with spaces") == "name_with_spaces"
        # Names starting with numbers get prefixed
        assert generator._sanitize_profile_name("123start") == "gaol_123start"

    def test_save_and_load_profile(self) -> None:
        """Test saving and loading profile."""
        with tempfile.TemporaryDirectory() as tmpdir:
            generator = AppArmorProfileGenerator(profiles_dir=Path(tmpdir))

            profile = generator.generate_default("test")
            saved_path = generator.save_profile(profile)

            assert saved_path.exists()
            content = saved_path.read_text()
            assert "profile test" in content

    def test_list_profiles(self) -> None:
        """Test listing profiles."""
        with tempfile.TemporaryDirectory() as tmpdir:
            generator = AppArmorProfileGenerator(profiles_dir=Path(tmpdir))

            # Initially empty
            assert len(generator.list_profiles()) == 0

            # Save some profiles
            for i in range(2):
                profile = generator.generate_default(f"profile_{i}")
                generator.save_profile(profile)

            profiles = generator.list_profiles()
            assert len(profiles) == 2

    def test_default_capabilities_defined(self) -> None:
        """Test that default capabilities are defined."""
        assert len(DEFAULT_CAPABILITIES_ALLOW) > 0
        assert "chown" in DEFAULT_CAPABILITIES_ALLOW
        assert len(DEFAULT_CAPABILITIES_DENY) > 0
        assert "sys_admin" in DEFAULT_CAPABILITIES_DENY

    def test_default_file_patterns_defined(self) -> None:
        """Test that default file patterns are defined."""
        assert len(DEFAULT_FILE_READ) > 0
        assert len(DEFAULT_FILE_DENY) > 0
        assert "/etc/shadow" in DEFAULT_FILE_DENY


# =============================================================================
# Integration Tests
# =============================================================================


class TestSecurityIntegration:
    """Integration tests for the security module."""

    def test_rule_roundtrip(self) -> None:
        """Test full rule creation and retrieval cycle."""
        with tempfile.TemporaryDirectory() as tmpdir:
            manager = RuleManager(
                rules_dir=Path(tmpdir),
                gaol_dir=Path(tmpdir) / "gaol",
            )

            # Create rule
            rule = SecurityRule(
                name="integration_test_rule",
                description="Integration test rule",
                condition="container and proc.name=bash",
                output="Shell spawned (user=%user.name container=%container.name)",
                priority=FalcoPriority.WARNING,
                tags=["container", "shell", "test"],
            )
            manager.create_rule(rule)

            # Retrieve and verify
            retrieved = manager.get_rule("integration_test_rule")
            assert retrieved.description == "Integration test rule"
            assert retrieved.priority == FalcoPriority.WARNING
            assert "container" in retrieved.tags

            # Update via disable/enable
            manager.disable_rule("integration_test_rule")
            updated = manager.get_rule("integration_test_rule")
            assert updated.enabled is False

            # Delete
            manager.delete_rule("integration_test_rule")
            assert len(manager.list_rules(include_builtin=False)) == 0

    def test_seccomp_profile_oci_compliance(self) -> None:
        """Test that generated seccomp profiles are OCI-compliant."""
        generator = SeccompProfileGenerator()
        profile = generator.generate_default("test")
        oci_json = profile.to_oci_json()

        # Check required OCI fields
        assert "defaultAction" in oci_json
        assert "architectures" in oci_json
        assert "syscalls" in oci_json

        # Validate JSON serialization
        json_str = json.dumps(oci_json)
        parsed = json.loads(json_str)
        assert parsed["defaultAction"] == "SCMP_ACT_ERRNO"

    def test_apparmor_profile_syntax(self) -> None:
        """Test that generated AppArmor profiles have valid syntax."""
        generator = AppArmorProfileGenerator()
        profile = generator.generate_default("test_container")
        text = profile.to_apparmor_text()

        # Check required syntax elements
        assert text.startswith("#")  # Comment header
        assert "profile test_container" in text
        assert "{" in text
        assert "}" in text
        assert "#include <abstractions/base>" in text

    def test_event_buffer_thread_safety(self) -> None:
        """Test event buffer is thread-safe."""
        import threading

        buffer = FalcoEventBuffer(max_size=1000)
        errors: list[Exception] = []

        def add_events(start: int, count: int) -> None:
            try:
                for i in range(count):
                    event = SecurityEvent(
                        uuid=f"thread-{start}-{i}",
                        timestamp=datetime.now(UTC),
                        rule="test",
                        priority=FalcoPriority.WARNING,
                        output="test",
                    )
                    buffer.add(event)
            except Exception as e:
                errors.append(e)

        threads = [
            threading.Thread(target=add_events, args=(i * 100, 100))
            for i in range(5)
        ]

        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(errors) == 0
        assert len(buffer) == 500



"""Core update execution logic for containers."""

from __future__ import annotations

import logging
import re
from datetime import datetime
from typing import TYPE_CHECKING, Callable

from gaol.distro import PackageManager, get_distro_info, get_package_manager_commands
from gaol.update.exceptions import (
    HealthCheckFailedError,
    PackageManagerError,
    UpdateExecutionError,
)
from gaol.update.models import (
    PackageUpdate,
    RollbackStrategy,
    UpdateCheck,
    UpdatePolicy,
    UpdateResult,
    UpdateStatus,
)

if TYPE_CHECKING:
    from gaol.runtimes.base import ContainerRuntime
    from gaol.store import StoredContainer


logger = logging.getLogger(__name__)


class UpdateExecutor:
    """Execute update operations on containers.

    Handles the low-level execution of package updates within containers,
    including checking for updates, applying them, and parsing results.

    Example:
        >>> executor = UpdateExecutor(runtime)
        >>> check = executor.check_updates(container_id, "mycontainer", "trixie")
        >>> if check.updates_available:
        ...     result = executor.apply_updates(container_id, stored, policy)
    """

    def __init__(self, runtime: ContainerRuntime) -> None:
        """Initialize the executor.

        Args:
            runtime: Container runtime to use for executing commands
        """
        self.runtime = runtime

    def check_updates(
        self,
        container_id: str,
        container_name: str,
        distro: str,
    ) -> UpdateCheck:
        """Check for available package updates.

        Args:
            container_id: Container ID
            container_name: Container name
            distro: Distribution name

        Returns:
            UpdateCheck with available updates
        """
        try:
            pm_commands = get_package_manager_commands(distro)
            distro_info = get_distro_info(distro)
            pm = distro_info.package_manager

            # Update package index
            exit_code, stdout, stderr = self.runtime.exec(
                container_id,
                ["sh", "-c", pm_commands["update"]],
            )

            if exit_code != 0 and pm != PackageManager.DNF:
                # DNF returns non-zero when updates are available
                return UpdateCheck(
                    container_name=container_name,
                    error_message=f"Failed to update package index: {stderr}",
                )

            # Get list of upgradable packages
            check_cmd = self._get_check_command(pm)
            exit_code, stdout, stderr = self.runtime.exec(
                container_id,
                ["sh", "-c", check_cmd],
            )

            packages = self._parse_upgradable_packages(stdout, pm)

            return UpdateCheck(
                container_name=container_name,
                updates_available=len(packages) > 0,
                package_count=len(packages),
                packages=packages,
            )

        except Exception as e:
            logger.error(f"Error checking updates for {container_name}: {e}")
            return UpdateCheck(
                container_name=container_name,
                error_message=str(e),
            )

    def apply_updates(
        self,
        container_id: str,
        stored: StoredContainer,
        policy: UpdatePolicy,
        snapshot_callback: Callable[[str], None] | None = None,
        health_check_callback: Callable[[str], bool] | None = None,
        rollback_callback: Callable[[str], None] | None = None,
    ) -> UpdateResult:
        """Apply updates to a container.

        Args:
            container_id: Container ID
            stored: Stored container info
            policy: Update policy
            snapshot_callback: Callback to create snapshot (name) -> None
            health_check_callback: Callback to run health check (container_id) -> bool
            rollback_callback: Callback to rollback (snapshot_name) -> None

        Returns:
            UpdateResult with operation outcome
        """
        result = UpdateResult(
            container_name=stored.config.name,
            status=UpdateStatus.CHECKING,
        )

        distro = stored.config.distro
        pm_commands = get_package_manager_commands(distro)
        distro_info = get_distro_info(distro)
        pm = distro_info.package_manager

        try:
            # Step 1: Create pre-update snapshot if requested
            if policy.pre_update_snapshot and snapshot_callback:
                result.status = UpdateStatus.DOWNLOADING
                snapshot_name = f"pre-update-{datetime.now().strftime('%Y%m%d-%H%M%S')}"
                try:
                    snapshot_callback(snapshot_name)
                    result.snapshot_name = snapshot_name
                    logger.info(f"Created pre-update snapshot: {snapshot_name}")
                except Exception as e:
                    logger.warning(f"Failed to create snapshot: {e}")
                    if policy.rollback_strategy == RollbackStrategy.AUTO:
                        raise UpdateExecutionError(
                            f"Snapshot required but failed: {e}"
                        )

            # Step 2: Update package index
            result.status = UpdateStatus.DOWNLOADING
            exit_code, stdout, stderr = self.runtime.exec(
                container_id,
                ["sh", "-c", pm_commands["update"]],
            )
            if exit_code != 0 and pm != PackageManager.DNF:
                raise PackageManagerError(pm.value, "update", stderr)

            # Step 3: Check what will be upgraded
            check_cmd = self._get_check_command(pm)
            exit_code, stdout, stderr = self.runtime.exec(
                container_id,
                ["sh", "-c", check_cmd],
            )
            packages_before = self._parse_upgradable_packages(stdout, pm)
            result.packages_checked = len(packages_before)

            if not packages_before:
                # No updates available
                result.status = UpdateStatus.COMPLETED
                result.completed_at = datetime.now()
                return result

            # Step 4: Apply upgrades
            result.status = UpdateStatus.APPLYING
            upgrade_cmd = self._get_upgrade_command(pm, policy.packages_to_skip)
            exit_code, stdout, stderr = self.runtime.exec(
                container_id,
                ["sh", "-c", upgrade_cmd],
            )

            if exit_code != 0:
                raise PackageManagerError(pm.value, "upgrade", stderr)

            # Parse upgraded packages from output
            result.packages_updated = packages_before  # Assume all were updated
            result.packages_upgraded = len(packages_before)

            # Step 5: Clean up
            self.runtime.exec(
                container_id,
                ["sh", "-c", pm_commands["clean"]],
            )

            # Step 6: Health check verification
            if policy.health_check_after and health_check_callback:
                result.status = UpdateStatus.VERIFYING
                try:
                    health_ok = health_check_callback(container_id)
                    result.health_check_passed = health_ok

                    if not health_ok:
                        raise HealthCheckFailedError(
                            stored.config.name,
                            "Container health check failed after update",
                        )

                except HealthCheckFailedError:
                    # Attempt rollback
                    if (
                        policy.rollback_strategy == RollbackStrategy.AUTO
                        and result.snapshot_name
                        and rollback_callback
                    ):
                        result.status = UpdateStatus.ROLLING_BACK
                        result.rollback_reason = "Health check failed after update"
                        try:
                            rollback_callback(result.snapshot_name)
                            result.rolled_back = True
                        except Exception as re:
                            logger.error(f"Rollback failed: {re}")
                            result.error_message = (
                                f"Health check failed and rollback failed: {re}"
                            )

                        result.status = UpdateStatus.FAILED
                        result.completed_at = datetime.now()
                        return result
                    else:
                        raise

            # Success
            result.status = UpdateStatus.COMPLETED
            result.completed_at = datetime.now()

        except PackageManagerError as e:
            result.status = UpdateStatus.FAILED
            result.completed_at = datetime.now()
            result.error_message = e.message
            result.error_output = e.stderr

            # Attempt rollback if configured
            if (
                policy.rollback_strategy == RollbackStrategy.AUTO
                and result.snapshot_name
                and rollback_callback
            ):
                result.status = UpdateStatus.ROLLING_BACK
                result.rollback_reason = str(e)
                try:
                    rollback_callback(result.snapshot_name)
                    result.rolled_back = True
                except Exception as re:
                    logger.error(f"Rollback failed: {re}")

                result.status = UpdateStatus.FAILED

        except Exception as e:
            result.status = UpdateStatus.FAILED
            result.completed_at = datetime.now()
            result.error_message = str(e)

        return result

    def _get_check_command(self, pm: PackageManager) -> str:
        """Get command to list upgradable packages."""
        commands = {
            PackageManager.APT: "apt list --upgradable 2>/dev/null | tail -n +2 || true",
            PackageManager.DNF: "dnf check-update 2>/dev/null | grep -v '^$' | grep -v '^Last' || true",
            PackageManager.APK: "apk version -l '<' 2>/dev/null || true",
            PackageManager.PACMAN: "pacman -Qu 2>/dev/null || true",
        }
        return commands.get(pm, "echo ''")

    def _get_upgrade_command(
        self, pm: PackageManager, skip_packages: list[str] | None = None
    ) -> str:
        """Get full upgrade command for a package manager."""
        base_commands = {
            PackageManager.APT: "DEBIAN_FRONTEND=noninteractive apt-get upgrade -y",
            PackageManager.DNF: "dnf upgrade -y",
            PackageManager.APK: "apk upgrade",
            PackageManager.PACMAN: "pacman -Syu --noconfirm",
        }

        cmd = base_commands.get(pm, "echo 'unsupported'")

        # Add package exclusions if specified
        if skip_packages:
            if pm == PackageManager.APT:
                # APT uses apt-mark hold for package exclusion
                hold_cmds = " && ".join(
                    f"apt-mark hold {pkg}" for pkg in skip_packages
                )
                unhold_cmds = " && ".join(
                    f"apt-mark unhold {pkg}" for pkg in skip_packages
                )
                cmd = f"{hold_cmds} && {cmd} && {unhold_cmds}"
            elif pm == PackageManager.DNF:
                excludes = " ".join(f"--exclude={pkg}" for pkg in skip_packages)
                cmd = f"dnf upgrade -y {excludes}"
            elif pm == PackageManager.PACMAN:
                ignores = " ".join(f"--ignore {pkg}" for pkg in skip_packages)
                cmd = f"pacman -Syu --noconfirm {ignores}"
            # APK doesn't have a simple exclude mechanism

        return cmd

    def _parse_upgradable_packages(
        self, output: str, pm: PackageManager
    ) -> list[PackageUpdate]:
        """Parse package manager output to find upgradable packages."""
        packages = []

        if not output.strip():
            return packages

        lines = output.strip().split("\n")

        for line in lines:
            line = line.strip()
            if not line:
                continue

            pkg = self._parse_package_line(line, pm)
            if pkg:
                packages.append(pkg)

        return packages

    def _parse_package_line(
        self, line: str, pm: PackageManager
    ) -> PackageUpdate | None:
        """Parse a single line of package output."""
        try:
            if pm == PackageManager.APT:
                # Format: package/release version [upgradable from: old_version]
                match = re.match(
                    r"^([^/]+)/\S+\s+(\S+)\s+.*\[upgradable from:\s*(\S+)\]",
                    line,
                )
                if match:
                    return PackageUpdate(
                        package_name=match.group(1),
                        new_version=match.group(2),
                        old_version=match.group(3),
                    )

            elif pm == PackageManager.DNF:
                # Format: package.arch version repo
                parts = line.split()
                if len(parts) >= 2:
                    pkg_arch = parts[0]
                    pkg_name = pkg_arch.rsplit(".", 1)[0] if "." in pkg_arch else pkg_arch
                    return PackageUpdate(
                        package_name=pkg_name,
                        new_version=parts[1] if len(parts) > 1 else None,
                    )

            elif pm == PackageManager.APK:
                # Format: package-version < new-version
                # Version starts with digits, e.g., vim-8.2.4000-r0 < 9.0.1000-r0
                match = re.match(r"^(.+)-(\d[\d.\w-]*)\s+<\s+(\S+)", line)
                if match:
                    return PackageUpdate(
                        package_name=match.group(1),
                        old_version=match.group(2),
                        new_version=match.group(3),
                    )

            elif pm == PackageManager.PACMAN:
                # Format: package old_version -> new_version
                match = re.match(r"^(\S+)\s+(\S+)\s+->\s+(\S+)", line)
                if match:
                    return PackageUpdate(
                        package_name=match.group(1),
                        old_version=match.group(2),
                        new_version=match.group(3),
                    )
                # Alternative format: package new_version
                parts = line.split()
                if len(parts) >= 2:
                    return PackageUpdate(
                        package_name=parts[0],
                        new_version=parts[1],
                    )

        except Exception as e:
            logger.debug(f"Failed to parse package line '{line}': {e}")

        return None



"""Pydantic models for service migration.

This module defines data structures for parsing systemd services,
discovering dependencies, and analyzing migration requirements.
"""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Any

from pydantic import BaseModel, Field


class UnitSection(BaseModel):
    """Parsed [Unit] section from a systemd service file."""

    description: str | None = None
    documentation: list[str] = Field(default_factory=list)
    after: list[str] = Field(default_factory=list)
    before: list[str] = Field(default_factory=list)
    wants: list[str] = Field(default_factory=list)
    requires: list[str] = Field(default_factory=list)
    conflicts: list[str] = Field(default_factory=list)
    binds_to: list[str] = Field(default_factory=list)
    part_of: list[str] = Field(default_factory=list)
    condition_path_exists: list[str] = Field(default_factory=list)


class ServiceSection(BaseModel):
    """Parsed [Service] section from a systemd service file."""

    type: str = "simple"
    exec_start: str | list[str] | None = None
    exec_start_pre: list[str] = Field(default_factory=list)
    exec_start_post: list[str] = Field(default_factory=list)
    exec_stop: str | list[str] | None = None
    exec_stop_post: list[str] = Field(default_factory=list)
    exec_reload: str | list[str] | None = None
    user: str = "root"
    group: str | None = None
    working_directory: str | None = None
    root_directory: str | None = None
    environment: dict[str, str] = Field(default_factory=dict)
    environment_file: list[str] = Field(default_factory=list)
    state_directory: list[str] = Field(default_factory=list)
    cache_directory: list[str] = Field(default_factory=list)
    logs_directory: list[str] = Field(default_factory=list)
    configuration_directory: list[str] = Field(default_factory=list)
    runtime_directory: list[str] = Field(default_factory=list)
    restart: str = "no"
    restart_sec: int | None = None
    timeout_start_sec: int | None = None
    timeout_stop_sec: int | None = None
    pid_file: str | None = None
    standard_output: str | None = None
    standard_error: str | None = None
    syslog_identifier: str | None = None
    # Security/capability related
    capability_bounding_set: list[str] = Field(default_factory=list)
    ambient_capabilities: list[str] = Field(default_factory=list)
    no_new_privileges: bool = False
    private_tmp: bool = False
    private_network: bool = False
    protect_system: str | None = None
    protect_home: str | None = None
    read_only_paths: list[str] = Field(default_factory=list)
    read_write_paths: list[str] = Field(default_factory=list)
    bind_paths: list[str] = Field(default_factory=list)
    bind_read_only_paths: list[str] = Field(default_factory=list)
    # Limit related
    limit_nofile: str | None = None
    limit_nproc: str | None = None
    limit_memlock: str | None = None
    # Other
    remain_after_exit: bool = False
    kill_mode: str | None = None
    kill_signal: str | None = None


class ServiceInstallSection(BaseModel):
    """Parsed [Install] section from a systemd service file."""

    wanted_by: list[str] = Field(default_factory=list)
    required_by: list[str] = Field(default_factory=list)
    alias: list[str] = Field(default_factory=list)
    also: list[str] = Field(default_factory=list)


class ParsedSystemdService(BaseModel):
    """Complete parsed systemd service file."""

    name: str
    """Service name (without .service suffix)."""

    file_path: Path
    """Path to the original service file."""

    unit: UnitSection = Field(default_factory=UnitSection)
    service: ServiceSection = Field(default_factory=ServiceSection)
    install: ServiceInstallSection = Field(default_factory=ServiceInstallSection)

    # Template service info
    is_template: bool = False
    """True if this is a template service (foo@.service)."""

    template_instance: str | None = None
    """Instance parameter for template services (e.g., 'instance' in foo@instance.service)."""

    # Drop-in files merged
    dropin_files: list[Path] = Field(default_factory=list)
    """List of drop-in configuration files that were merged."""

    # Raw sections for any non-standard keys
    raw_sections: dict[str, dict[str, Any]] = Field(default_factory=dict)

    def get_binary_path(self) -> str | None:
        """Extract the binary path from ExecStart."""
        exec_start = self.service.exec_start
        if not exec_start:
            return None
        # Handle list format (multiple ExecStart lines)
        if isinstance(exec_start, list):
            exec_start = exec_start[0] if exec_start else None
        if not exec_start:
            return None
        # Handle various prefix characters: -, @, +, !, !!
        cmd = exec_start.lstrip("-@+!").strip()
        # Split on whitespace to get the binary
        parts = cmd.split()
        if not parts:
            return None
        return parts[0]


class DiscoveredDependencies(BaseModel):
    """Dependencies discovered from analyzing a service."""

    # Package dependencies
    packages: list[str] = Field(default_factory=list)
    """Packages required by the service binary and dependencies."""

    binary_package_map: dict[str, str] = Field(default_factory=dict)
    """Mapping of binary paths to their packages."""

    # File dependencies
    config_files: list[str] = Field(default_factory=list)
    """Configuration files needed by the service."""

    config_directories: list[str] = Field(default_factory=list)
    """Configuration directories needed by the service."""

    # Data and state
    data_directories: list[str] = Field(default_factory=list)
    """Data directories that should be mounted or copied."""

    state_directories: list[str] = Field(default_factory=list)
    """State directories (typically in /var/lib)."""

    # Network
    listening_ports: list[int] = Field(default_factory=list)
    """TCP/UDP ports the service listens on."""

    # Related services
    socket_units: list[str] = Field(default_factory=list)
    """Associated .socket units for socket activation."""

    timer_units: list[str] = Field(default_factory=list)
    """Associated .timer units."""

    # Capabilities and privileges
    required_capabilities: list[str] = Field(default_factory=list)
    """Linux capabilities required by the service."""


class MigrationWarningSeverity(str, Enum):
    """Severity level for migration warnings."""

    INFO = "info"
    WARNING = "warning"
    CRITICAL = "critical"


class MigrationWarning(BaseModel):
    """Warning about potential issues during migration."""

    category: str
    """Category of warning (e.g., 'socket_activation', 'kernel_access')."""

    message: str
    """Human-readable warning message."""

    severity: MigrationWarningSeverity = MigrationWarningSeverity.WARNING
    """Severity level."""

    suggestion: str | None = None
    """Suggested action to address the warning."""

    def __str__(self) -> str:
        """Format warning for display."""
        prefix = {
            MigrationWarningSeverity.INFO: "INFO",
            MigrationWarningSeverity.WARNING: "WARN",
            MigrationWarningSeverity.CRITICAL: "CRITICAL",
        }[self.severity]
        msg = f"[{prefix}] {self.category}: {self.message}"
        if self.suggestion:
            msg += f"\n  Suggestion: {self.suggestion}"
        return msg


class MigrationAnalysis(BaseModel):
    """Complete analysis of a service migration."""

    service: ParsedSystemdService
    """The parsed systemd service."""

    dependencies: DiscoveredDependencies
    """Discovered dependencies."""

    warnings: list[MigrationWarning] = Field(default_factory=list)
    """Warnings about migration issues."""

    # Computed recommendations
    recommended_runtime: str = "nspawn"
    """Recommended container runtime."""

    recommended_distro: str = "trixie"
    """Recommended base distribution."""

    requires_privileged: bool = False
    """Whether the container needs privileged mode."""

    requires_host_network: bool = False
    """Whether the service needs host network access."""

    can_migrate: bool = True
    """Whether the service can be safely migrated."""

    migration_blockers: list[str] = Field(default_factory=list)
    """Reasons that prevent migration (if can_migrate is False)."""

    @property
    def has_critical_warnings(self) -> bool:
        """Check if there are any critical warnings."""
        return any(w.severity == MigrationWarningSeverity.CRITICAL for w in self.warnings)


class WellKnownService(BaseModel):
    """Pre-configured knowledge about a well-known service."""

    name: str
    """Service name."""

    packages: list[str] = Field(default_factory=list)
    """Required packages."""

    config_paths: list[str] = Field(default_factory=list)
    """Configuration file/directory paths."""

    data_paths: list[str] = Field(default_factory=list)
    """Data directory paths."""

    ports: list[int] = Field(default_factory=list)
    """Default listening ports."""

    notes: str | None = None
    """Additional migration notes."""


# Well-known services registry
WELLKNOWN_SERVICES: dict[str, WellKnownService] = {
    "nginx": WellKnownService(
        name="nginx",
        packages=["nginx"],
        config_paths=["/etc/nginx/"],
        data_paths=["/var/www/", "/usr/share/nginx/html/"],
        ports=[80, 443],
    ),
    "apache2": WellKnownService(
        name="apache2",
        packages=["apache2"],
        config_paths=["/etc/apache2/"],
        data_paths=["/var/www/"],
        ports=[80, 443],
    ),
    "postgresql": WellKnownService(
        name="postgresql",
        packages=["postgresql"],
        config_paths=["/etc/postgresql/"],
        data_paths=["/var/lib/postgresql/"],
        ports=[5432],
        notes="Database files should be migrated carefully with pg_dump",
    ),
    "mysql": WellKnownService(
        name="mysql",
        packages=["mysql-server"],
        config_paths=["/etc/mysql/"],
        data_paths=["/var/lib/mysql/"],
        ports=[3306],
        notes="Use mysqldump for data migration",
    ),
    "mariadb": WellKnownService(
        name="mariadb",
        packages=["mariadb-server"],
        config_paths=["/etc/mysql/"],
        data_paths=["/var/lib/mysql/"],
        ports=[3306],
        notes="Use mysqldump for data migration",
    ),
    "redis-server": WellKnownService(
        name="redis-server",
        packages=["redis-server"],
        config_paths=["/etc/redis/"],
        data_paths=["/var/lib/redis/"],
        ports=[6379],
    ),
    "memcached": WellKnownService(
        name="memcached",
        packages=["memcached"],
        config_paths=["/etc/memcached.conf"],
        data_paths=[],
        ports=[11211],
    ),
    "sshd": WellKnownService(
        name="sshd",
        packages=["openssh-server"],
        config_paths=["/etc/ssh/"],
        data_paths=[],
        ports=[22],
        notes="Container SSH requires proper key management",
    ),
    "ssh": WellKnownService(
        name="ssh",
        packages=["openssh-server"],
        config_paths=["/etc/ssh/"],
        data_paths=[],
        ports=[22],
        notes="Container SSH requires proper key management",
    ),
    "docker": WellKnownService(
        name="docker",
        packages=["docker.io"],
        config_paths=["/etc/docker/"],
        data_paths=["/var/lib/docker/"],
        ports=[2375, 2376],
        notes="Docker-in-container requires privileged mode",
    ),
    "containerd": WellKnownService(
        name="containerd",
        packages=["containerd"],
        config_paths=["/etc/containerd/"],
        data_paths=["/var/lib/containerd/"],
        ports=[],
    ),
    "haproxy": WellKnownService(
        name="haproxy",
        packages=["haproxy"],
        config_paths=["/etc/haproxy/"],
        data_paths=[],
        ports=[80, 443, 8080],
    ),
    "bind9": WellKnownService(
        name="bind9",
        packages=["bind9"],
        config_paths=["/etc/bind/"],
        data_paths=["/var/cache/bind/"],
        ports=[53],
    ),
    "named": WellKnownService(
        name="named",
        packages=["bind9"],
        config_paths=["/etc/bind/"],
        data_paths=["/var/cache/bind/"],
        ports=[53],
    ),
    "postfix": WellKnownService(
        name="postfix",
        packages=["postfix"],
        config_paths=["/etc/postfix/"],
        data_paths=["/var/spool/postfix/", "/var/mail/"],
        ports=[25, 587],
    ),
    "dovecot": WellKnownService(
        name="dovecot",
        packages=["dovecot-core", "dovecot-imapd"],
        config_paths=["/etc/dovecot/"],
        data_paths=["/var/mail/"],
        ports=[143, 993, 110, 995],
    ),
    "grafana-server": WellKnownService(
        name="grafana-server",
        packages=["grafana"],
        config_paths=["/etc/grafana/"],
        data_paths=["/var/lib/grafana/"],
        ports=[3000],
    ),
    "prometheus": WellKnownService(
        name="prometheus",
        packages=["prometheus"],
        config_paths=["/etc/prometheus/"],
        data_paths=["/var/lib/prometheus/"],
        ports=[9090],
    ),
    "elasticsearch": WellKnownService(
        name="elasticsearch",
        packages=["elasticsearch"],
        config_paths=["/etc/elasticsearch/"],
        data_paths=["/var/lib/elasticsearch/"],
        ports=[9200, 9300],
    ),
    "rabbitmq-server": WellKnownService(
        name="rabbitmq-server",
        packages=["rabbitmq-server"],
        config_paths=["/etc/rabbitmq/"],
        data_paths=["/var/lib/rabbitmq/"],
        ports=[5672, 15672],
    ),
    "mongod": WellKnownService(
        name="mongod",
        packages=["mongodb-org-server"],
        config_paths=["/etc/mongod.conf"],
        data_paths=["/var/lib/mongodb/"],
        ports=[27017],
    ),
    "influxdb": WellKnownService(
        name="influxdb",
        packages=["influxdb"],
        config_paths=["/etc/influxdb/"],
        data_paths=["/var/lib/influxdb/"],
        ports=[8086],
    ),
    "jenkins": WellKnownService(
        name="jenkins",
        packages=["jenkins"],
        config_paths=["/etc/default/jenkins"],
        data_paths=["/var/lib/jenkins/"],
        ports=[8080],
    ),
    "gitlab-runsvdir": WellKnownService(
        name="gitlab-runsvdir",
        packages=["gitlab-ce"],
        config_paths=["/etc/gitlab/"],
        data_paths=["/var/opt/gitlab/"],
        ports=[80, 443, 22],
    ),
    "cups": WellKnownService(
        name="cups",
        packages=["cups"],
        config_paths=["/etc/cups/"],
        data_paths=["/var/spool/cups/"],
        ports=[631],
    ),
    "openvpn": WellKnownService(
        name="openvpn",
        packages=["openvpn"],
        config_paths=["/etc/openvpn/"],
        data_paths=[],
        ports=[1194],
        notes="Requires TUN/TAP device access",
    ),
    "wireguard": WellKnownService(
        name="wireguard",
        packages=["wireguard-tools"],
        config_paths=["/etc/wireguard/"],
        data_paths=[],
        ports=[51820],
        notes="Requires kernel module and NET_ADMIN capability",
    ),
    "tor": WellKnownService(
        name="tor",
        packages=["tor"],
        config_paths=["/etc/tor/"],
        data_paths=["/var/lib/tor/"],
        ports=[9050, 9051],
    ),
    "squid": WellKnownService(
        name="squid",
        packages=["squid"],
        config_paths=["/etc/squid/"],
        data_paths=["/var/spool/squid/"],
        ports=[3128],
    ),
    "unbound": WellKnownService(
        name="unbound",
        packages=["unbound"],
        config_paths=["/etc/unbound/"],
        data_paths=[],
        ports=[53],
    ),
    "dnsmasq": WellKnownService(
        name="dnsmasq",
        packages=["dnsmasq"],
        config_paths=["/etc/dnsmasq.conf", "/etc/dnsmasq.d/"],
        data_paths=[],
        ports=[53, 67],
    ),
    "fail2ban": WellKnownService(
        name="fail2ban",
        packages=["fail2ban"],
        config_paths=["/etc/fail2ban/"],
        data_paths=["/var/lib/fail2ban/"],
        ports=[],
        notes="Requires access to host logs and iptables",
    ),
    "jellyfin": WellKnownService(
        name="jellyfin",
        packages=["jellyfin", "jellyfin-web", "jellyfin-ffmpeg7"],
        config_paths=["/etc/jellyfin/", "/etc/default/jellyfin"],
        data_paths=["/var/lib/jellyfin/", "/var/cache/jellyfin/", "/var/log/jellyfin/"],
        ports=[8096, 8920],
        notes="Requires ffmpeg symlinks: ln -sf /usr/lib/jellyfin-ffmpeg/ffmpeg /usr/bin/ffmpeg",
    ),
}



"""Tests for gaol models."""

import pytest
from pydantic import ValidationError

from gaol.models.container import (
    ContainerConfig,
    ContainerState,
    ContainerStats,
    ContainerStatus,
    HealthCheck,
    HealthCheckResult,
    HealthStatus,
    NetworkConfig,
    NetworkMode,
    ResourceLimits,
    Snapshot,
)
from gaol.models.profile import Profile, ProfileDependency


class TestContainerConfig:
    """Tests for ContainerConfig model."""

    def test_minimal_config(self) -> None:
        """Test creating a container config with minimal options."""
        config = ContainerConfig(name="test-container")
        assert config.name == "test-container"
        assert config.runtime == "docker"
        assert config.distro == "trixie"
        assert config.profiles == []
        assert config.network.mode == NetworkMode.SHARED

    def test_full_config(self) -> None:
        """Test creating a container config with all options."""
        config = ContainerConfig(
            name="full-container",
            runtime="docker",
            distro="bookworm",
            image="debian:bookworm-slim",
            profiles=["base", "tailscale"],
            network=NetworkConfig(
                mode=NetworkMode.BRIDGED,
                ports={8080: 80, 8443: 443},
                dns=["8.8.8.8", "8.8.4.4"],
                hostname="myhost",
            ),
            resources=ResourceLimits(
                memory_mb=512,
                cpu_count=2.0,
            ),
            volumes={"/data": "/container/data"},
            environment={"DEBUG": "1"},
            working_dir="/app",
            privileged=False,
        )
        assert config.name == "full-container"
        assert config.image == "debian:bookworm-slim"
        assert config.network.mode == NetworkMode.BRIDGED
        assert config.network.ports[8080] == 80
        assert config.resources.memory_mb == 512

    def test_invalid_name(self) -> None:
        """Test that invalid container names are rejected."""
        with pytest.raises(ValidationError):
            ContainerConfig(name="")  # Too short

        with pytest.raises(ValidationError):
            ContainerConfig(name="-invalid")  # Can't start with hyphen

        with pytest.raises(ValidationError):
            ContainerConfig(name="has spaces")  # No spaces allowed


class TestContainerState:
    """Tests for ContainerState model."""

    def test_state_creation(self) -> None:
        """Test creating a container state."""
        state = ContainerState(
            id="abc123",
            name="test-container",
            runtime="docker",
            status=ContainerStatus.RUNNING,
            image="debian:trixie",
        )
        assert state.id == "abc123"
        assert state.status == ContainerStatus.RUNNING
        assert state.ip_address is None

    def test_state_with_network(self) -> None:
        """Test container state with network info."""
        state = ContainerState(
            id="abc123",
            name="test-container",
            runtime="docker",
            status=ContainerStatus.RUNNING,
            image="debian:trixie",
            ip_address="172.17.0.2",
            ports={8080: 80},
        )
        assert state.ip_address == "172.17.0.2"
        assert state.ports[8080] == 80


class TestProfile:
    """Tests for Profile model."""

    def test_minimal_profile(self) -> None:
        """Test creating a minimal profile."""
        profile = Profile(name="test")
        assert profile.name == "test"
        assert profile.packages == []
        assert profile.services == []

    def test_full_profile(self) -> None:
        """Test creating a full profile."""
        profile = Profile(
            name="tailscale",
            description="Tailscale VPN",
            requires_root=True,
            packages=["tailscale"],
            services=["tailscaled"],
            setup_commands=["tailscale up"],
            environment={"TAILSCALE_AUTHKEY": ""},
            exposed_ports=[41641],
        )
        assert profile.requires_root is True
        assert "tailscale" in profile.packages
        assert 41641 in profile.exposed_ports

    def test_profile_from_yaml(self) -> None:
        """Test creating a profile from YAML-like dict."""
        data = {
            "name": "base",
            "description": "Base profile",
            "packages": ["curl", "wget"],
            "depends_on": ["minimal", {"name": "optional-dep", "optional": True}],
        }
        profile = Profile.from_yaml(data)
        assert profile.name == "base"
        assert len(profile.depends_on) == 2
        assert profile.depends_on[0].name == "minimal"
        assert profile.depends_on[0].optional is False
        assert profile.depends_on[1].name == "optional-dep"
        assert profile.depends_on[1].optional is True


class TestProfileDependency:
    """Tests for ProfileDependency model."""

    def test_required_dependency(self) -> None:
        """Test creating a required dependency."""
        dep = ProfileDependency(name="base")
        assert dep.name == "base"
        assert dep.optional is False

    def test_optional_dependency(self) -> None:
        """Test creating an optional dependency."""
        dep = ProfileDependency(name="gui", optional=True)
        assert dep.name == "gui"
        assert dep.optional is True


class TestSnapshot:
    """Tests for Snapshot model."""

    def test_snapshot_creation(self) -> None:
        """Test creating a snapshot."""
        from datetime import datetime

        snapshot = Snapshot(
            id="container1-snap1",
            name="snap1",
            container_id="container1",
            container_name="my-container",
            created_at=datetime.now(),
        )
        assert snapshot.id == "container1-snap1"
        assert snapshot.name == "snap1"
        assert snapshot.container_id == "container1"
        assert snapshot.container_name == "my-container"
        assert snapshot.size_bytes is None
        assert snapshot.description is None

    def test_snapshot_with_all_fields(self) -> None:
        """Test creating a snapshot with all fields."""
        from datetime import datetime

        created = datetime(2024, 1, 15, 10, 30, 0)
        snapshot = Snapshot(
            id="container1-backup",
            name="backup",
            container_id="container1",
            container_name="my-container",
            created_at=created,
            size_bytes=1024 * 1024 * 100,  # 100MB
            description="Pre-upgrade backup",
        )
        assert snapshot.size_bytes == 104857600
        assert snapshot.description == "Pre-upgrade backup"
        assert snapshot.created_at == created


class TestContainerStats:
    """Tests for ContainerStats model."""

    def test_stats_creation(self) -> None:
        """Test creating container stats."""
        stats = ContainerStats(
            container_id="abc123",
            container_name="test-container",
            cpu_percent=25.5,
            cpu_count=4,
            memory_used_bytes=512 * 1024 * 1024,  # 512 MB
            memory_limit_bytes=1024 * 1024 * 1024,  # 1 GB
            memory_percent=50.0,
            network_rx_bytes=1000000,
            network_tx_bytes=500000,
            block_read_bytes=2000000,
            block_write_bytes=1000000,
            pids=10,
        )
        assert stats.container_id == "abc123"
        assert stats.cpu_percent == 25.5
        assert stats.memory_used_mb == 512.0
        assert stats.memory_limit_mb == 1024.0
        assert stats.pids == 10

    def test_stats_defaults(self) -> None:
        """Test container stats defaults."""
        stats = ContainerStats(
            container_id="abc123",
            container_name="test-container",
        )
        assert stats.cpu_percent == 0.0
        assert stats.memory_used_bytes == 0
        assert stats.memory_limit_bytes is None
        assert stats.memory_limit_mb is None
        assert stats.network_rx_bytes == 0
        assert stats.pids == 0


class TestHealthCheck:
    """Tests for HealthCheck model."""

    def test_health_check_creation(self) -> None:
        """Test creating a health check."""
        check = HealthCheck(
            command=["curl", "-f", "http://localhost/health"],
            interval_seconds=30,
            timeout_seconds=10,
            retries=3,
        )
        assert check.command == ["curl", "-f", "http://localhost/health"]
        assert check.interval_seconds == 30
        assert check.timeout_seconds == 10
        assert check.retries == 3
        assert check.start_period_seconds == 0

    def test_health_check_defaults(self) -> None:
        """Test health check defaults."""
        check = HealthCheck(command=["true"])
        assert check.interval_seconds == 30
        assert check.timeout_seconds == 10
        assert check.retries == 3
        assert check.start_period_seconds == 0


class TestHealthCheckResult:
    """Tests for HealthCheckResult model."""

    def test_healthy_result(self) -> None:
        """Test healthy check result."""
        result = HealthCheckResult(
            status=HealthStatus.HEALTHY,
            exit_code=0,
            output="OK",
        )
        assert result.status == HealthStatus.HEALTHY
        assert result.exit_code == 0
        assert result.consecutive_failures == 0

    def test_unhealthy_result(self) -> None:
        """Test unhealthy check result."""
        result = HealthCheckResult(
            status=HealthStatus.UNHEALTHY,
            exit_code=1,
            output="Connection refused",
            consecutive_failures=3,
        )
        assert result.status == HealthStatus.UNHEALTHY
        assert result.exit_code == 1
        assert result.consecutive_failures == 3

    def test_health_status_values(self) -> None:
        """Test health status enum values."""
        assert HealthStatus.HEALTHY.value == "healthy"
        assert HealthStatus.UNHEALTHY.value == "unhealthy"
        assert HealthStatus.STARTING.value == "starting"
        assert HealthStatus.UNKNOWN.value == "unknown"



"""Exceptions for Copy-on-Write filesystem operations."""

from __future__ import annotations


class CoWError(Exception):
    """Base exception for CoW operations."""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class PoolNotFoundError(CoWError):
    """Pool does not exist."""

    def __init__(self, name: str) -> None:
        self.pool_name = name
        super().__init__(f"Pool not found: {name}")


class PoolExistsError(CoWError):
    """Pool already exists."""

    def __init__(self, name: str) -> None:
        self.pool_name = name
        super().__init__(f"Pool already exists: {name}")


class PoolNotEmptyError(CoWError):
    """Pool has datasets and cannot be destroyed."""

    def __init__(self, name: str, dataset_count: int) -> None:
        self.pool_name = name
        self.dataset_count = dataset_count
        super().__init__(
            f"Pool '{name}' has {dataset_count} datasets. Use --force to destroy."
        )


class DatasetNotFoundError(CoWError):
    """Dataset does not exist."""

    def __init__(self, pool: str, name: str) -> None:
        self.pool = pool
        self.dataset_name = name
        super().__init__(f"Dataset not found: {pool}/{name}")


class DatasetExistsError(CoWError):
    """Dataset already exists."""

    def __init__(self, pool: str, name: str) -> None:
        self.pool = pool
        self.dataset_name = name
        super().__init__(f"Dataset already exists: {pool}/{name}")


class DatasetNotEmptyError(CoWError):
    """Dataset has snapshots and cannot be destroyed."""

    def __init__(self, pool: str, name: str, snapshot_count: int) -> None:
        self.pool = pool
        self.dataset_name = name
        self.snapshot_count = snapshot_count
        super().__init__(
            f"Dataset '{pool}/{name}' has {snapshot_count} snapshots. Use --force to destroy."
        )


class SnapshotNotFoundError(CoWError):
    """Snapshot does not exist."""

    def __init__(self, pool: str, dataset: str, name: str) -> None:
        self.pool = pool
        self.dataset = dataset
        self.snapshot_name = name
        super().__init__(f"Snapshot not found: {pool}/{dataset}@{name}")


class SnapshotExistsError(CoWError):
    """Snapshot already exists."""

    def __init__(self, pool: str, dataset: str, name: str) -> None:
        self.pool = pool
        self.dataset = dataset
        self.snapshot_name = name
        super().__init__(f"Snapshot already exists: {pool}/{dataset}@{name}")


class SnapshotHasClonesError(CoWError):
    """Snapshot has clones and cannot be deleted."""

    def __init__(self, pool: str, dataset: str, name: str) -> None:
        self.pool = pool
        self.dataset = dataset
        self.snapshot_name = name
        super().__init__(
            f"Snapshot '{pool}/{dataset}@{name}' has dependent clones. "
            "Promote or destroy clones first."
        )


class CloneNotFoundError(CoWError):
    """Clone does not exist."""

    def __init__(self, pool: str, name: str) -> None:
        self.pool = pool
        self.clone_name = name
        super().__init__(f"Clone not found: {pool}/{name}")


class BackendNotAvailableError(CoWError):
    """CoW backend is not available on this system."""

    def __init__(self, backend: str, reason: str | None = None) -> None:
        self.backend = backend
        self.reason = reason
        msg = f"Backend '{backend}' is not available"
        if reason:
            msg += f": {reason}"
        super().__init__(msg)


class InvalidMountPointError(CoWError):
    """Mount point is not a valid CoW filesystem."""

    def __init__(self, path: str, expected_type: str) -> None:
        self.path = path
        self.expected_type = expected_type
        super().__init__(
            f"Path '{path}' is not a valid {expected_type} filesystem mount point"
        )


class InsufficientPermissionsError(CoWError):
    """Operation requires elevated permissions."""

    def __init__(self, operation: str) -> None:
        self.operation = operation
        super().__init__(f"Operation '{operation}' requires root privileges")


class SendReceiveError(CoWError):
    """Error during send/receive operation."""

    def __init__(self, operation: str, reason: str) -> None:
        self.operation = operation
        self.reason = reason
        super().__init__(f"Send/receive {operation} failed: {reason}")


class QuotaError(CoWError):
    """Quota-related error."""

    def __init__(self, message: str) -> None:
        super().__init__(message)



"""Network manager coordinating all networking operations."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gaol.networking.bridge import BridgeConfig, ensure_bridge, get_bridge_info
from gaol.networking.dns import generate_resolv_conf, get_default_dns_config, get_system_dns
from gaol.networking.firewall import FirewallManager, ForwardingRule
from gaol.networking.models import DNSConfig, NetworkConfigV2
from gaol.networking.ports import PortError, validate_port_mappings

if TYPE_CHECKING:
    from gaol.runtimes.base import ContainerRuntime


class NetworkManager:
    """Manages networking for containers and VMs."""

    def __init__(self, use_sudo: bool = True) -> None:
        self.use_sudo = use_sudo
        self._firewall: FirewallManager | None = None

    @property
    def firewall(self) -> FirewallManager:
        """Get or create the firewall manager."""
        if self._firewall is None:
            self._firewall = FirewallManager(use_sudo=self.use_sudo)
            self._firewall.setup()
        return self._firewall

    def validate_network_config(
        self,
        config: NetworkConfigV2,
        runtime: ContainerRuntime | None = None,
        container_name: str | None = None,
    ) -> list[PortError]:
        """Validate a network configuration.

        Args:
            config: Network configuration to validate
            runtime: Runtime for conflict checking
            container_name: Container name (for excluding from conflict check)

        Returns:
            List of errors found (empty if valid)
        """
        errors: list[PortError] = []

        # Validate port mappings
        errors.extend(
            validate_port_mappings(
                list(config.ports.mappings),
                check_availability=True,
                check_conflicts=True,
                runtime=runtime,
                exclude_container=container_name,
            )
        )

        return errors

    def setup_container_network(
        self,
        container_id: str,
        config: NetworkConfigV2,
        guest_ip: str | None = None,
    ) -> None:
        """Set up networking for a container.

        Args:
            container_id: Container ID/name
            config: Network configuration
            guest_ip: Guest IP (required for libvirt port forwarding)
        """
        # For libvirt VMs with port mappings, set up firewall rules
        if guest_ip and config.ports.mappings:
            for mapping in config.ports.mappings:
                self.firewall.add_forwarding_rule(
                    ForwardingRule(
                        host_port=mapping.host_port,
                        guest_ip=guest_ip,
                        guest_port=mapping.container_port,
                        protocol=mapping.protocol,
                        host_ip=mapping.host_ip,
                        comment=f"gaol:{container_id}",
                    )
                )

    def teardown_container_network(
        self,
        _container_id: str,
        _config: NetworkConfigV2,
        guest_ip: str | None = None,
    ) -> None:
        """Tear down networking for a container.

        Args:
            _container_id: Container ID/name (reserved for future use)
            _config: Network configuration (reserved for future use)
            guest_ip: Guest IP (for firewall rule cleanup)
        """
        if guest_ip:
            self.firewall.clear_rules_for_vm(guest_ip)

    def get_dns_for_container(
        self,
        config: NetworkConfigV2,
        use_system_dns: bool = True,
    ) -> DNSConfig:
        """Get DNS configuration for a container.

        Args:
            config: Network configuration
            use_system_dns: Fall back to system DNS if not configured

        Returns:
            DNSConfig to use
        """
        if config.dns.nameservers:
            return config.dns

        if use_system_dns:
            return get_system_dns()

        # Return default public DNS
        return get_default_dns_config()

    def generate_resolv_conf_for_container(
        self,
        config: NetworkConfigV2,
        use_system_dns: bool = True,
    ) -> str:
        """Generate resolv.conf content for a container.

        Args:
            config: Network configuration
            use_system_dns: Fall back to system DNS if not configured

        Returns:
            resolv.conf content
        """
        dns_config = self.get_dns_for_container(config, use_system_dns)
        return generate_resolv_conf(dns_config)

    def ensure_bridge_for_mode(
        self,
        network_mode: str,
        bridge_config: BridgeConfig | None = None,
    ) -> str | None:
        """Ensure a bridge exists for bridged networking mode.

        Args:
            network_mode: Network mode (BRIDGED, etc.)
            bridge_config: Custom bridge configuration

        Returns:
            Bridge name if one was ensured, None otherwise
        """
        if network_mode.lower() != "bridged":
            return None

        config = bridge_config or BridgeConfig(name="br0")

        info = get_bridge_info(config.name)
        if not info:
            ensure_bridge(config, self.use_sudo)

        return config.name


# Singleton instance
_network_manager: NetworkManager | None = None


def get_network_manager() -> NetworkManager:
    """Get the global network manager instance."""
    global _network_manager
    if _network_manager is None:
        _network_manager = NetworkManager()
    return _network_manager



"""Exceptions for the sharing module."""

from __future__ import annotations


class SharingError(Exception):
    """Base exception for sharing operations."""

    pass


class ImportError(SharingError):
    """Error during content import."""

    def __init__(
        self,
        url: str,
        reason: str,
        cause: Exception | None = None,
    ) -> None:
        super().__init__(f"Failed to import from {url}: {reason}")
        self.url = url
        self.reason = reason
        self.cause = cause


class FetchError(SharingError):
    """Error fetching content from URL."""

    def __init__(
        self,
        url: str,
        reason: str,
        cause: Exception | None = None,
    ) -> None:
        super().__init__(f"Failed to fetch {url}: {reason}")
        self.url = url
        self.reason = reason
        self.cause = cause


class ContentValidationError(SharingError):
    """Error validating imported content."""

    def __init__(
        self,
        content_type: str,
        reason: str,
        errors: list[str] | None = None,
    ) -> None:
        msg = f"Invalid {content_type}: {reason}"
        if errors:
            msg += f" ({len(errors)} errors)"
        super().__init__(msg)
        self.content_type = content_type
        self.reason = reason
        self.errors = errors or []


class ContentTypeDetectionError(SharingError):
    """Could not auto-detect content type."""

    def __init__(self, url: str) -> None:
        super().__init__(
            f"Could not detect content type for {url}. "
            "Use --type to specify profile or spec."
        )
        self.url = url


class SecurityScanError(SharingError):
    """Security scan found critical issues."""

    def __init__(
        self,
        url: str,
        findings: list[str],
    ) -> None:
        msg = f"Security scan failed for {url}: {len(findings)} critical issue(s) found"
        super().__init__(msg)
        self.url = url
        self.findings = findings


class ProfileExistsError(SharingError):
    """Profile already exists in store."""

    def __init__(self, name: str, path: str) -> None:
        super().__init__(f"Profile already exists: {name} (at {path})")
        self.name = name
        self.path = path


class SpecExistsError(SharingError):
    """Spec already exists in store."""

    def __init__(self, name: str, path: str) -> None:
        super().__init__(f"Spec already exists: {name} (at {path})")
        self.name = name
        self.path = path


class URLResolveError(SharingError):
    """Error resolving URL reference."""

    def __init__(self, reference: str, reason: str) -> None:
        super().__init__(f"Could not resolve '{reference}': {reason}")
        self.reference = reference
        self.reason = reason
